<?php class apikey {

    protected static $fields = array(
        'name' => validator::ANY, //actually required but is auto generated
        'keyhash' => validator::REQUIRED,
        'iplowest' => array(validator::REQUIRED, validator::IP4ADDRESS),
        'iphighest' => array(validator::IP4ADDRESS),
    );

    protected static $helpTexts = array();

    private static $hashAlgorithm=PASSWORD_BCRYPT;

    /**
     * @param int $id
     * @return array
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws ServerException
     */
    public static function getById($id){
        session::requireAdmin();
        $ret=database::queryGetOne(
            'SELECT id, scope, iplowest, iphighest FROM apikey WHERE id=:id',
            array(':id'=>$id)
        );
        if($ret){
            $ret['iplowest']=long2ip($ret['iplowest']);
            $ret['iphighest']=long2ip($ret['iphighest']);
        }
        return $ret;
    }

    /**
     * @param string $name
     * @return array|void
     * @throws ForbiddenException
     * @throws ServerException
     */
    public static function getByName($name){
        session::requireAdmin();
        throw new ServerException('getByName not supported');
    }

    /**
     * @param array $request
     * @return array
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws ServerException
     */
    public static function getAll($request = array()){
        session::requireAdmin();
        $ret=database::queryGetAll(
            'SELECT SQL_CALC_FOUND_ROWS id, scope, iplowest, iphighest FROM apikey',
            $request
        );
        if($ret && isset($ret['rows'])){
            foreach ($ret['rows'] as &$row){
                $row['iplowest']=long2ip($row['iplowest']);
                $row['iphighest']=long2ip($row['iphighest']);
            }
        }
        $ret['scopes']=static::getScopes();
        return $ret;
    }

    /**
     * @param string $key
     * @param string $value
     * @param array $request
     * @return void
     * @throws ForbiddenException
     * @throws ServerException
     */
    public static function getByProperty($key, $value, $request=array()){
        session::requireAdmin();
        throw new ServerException('getByProperty not supported');
    }

    /**
     * @param array $keyValuePairs
     * @param array $request
     * @return void
     * @throws ForbiddenException
     * @throws ServerException
     */
    public static function getByProperties($keyValuePairs, $request=array()){
        session::requireAdmin();
        throw new ServerException('getByProperties not supported');
    }

    /**
     * @param $id
     * @param array $request
     * @return array|array[]
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws ServerException
     * @throws NotFoundException
     */
    public static function update($id, $request=array()){
        session::requireAdmin();
        if(isset($request['keyhash'])){
            baseobject::beforeUpdateHook($id, $request, 'apikey');
            $result=static::regenerateKey($id);
            baseobject::afterUpdateHook($id, 'apikey');
            return array('updated'=>$result);
        }
        $key=static::getById($id);
        baseobject::beforeUpdateHook($id, $request, 'apikey');
        foreach(static::$fields as $field=>$validations){
            if(isset($request[$field])){
                validator::validate($field,$request[$field],$validations); //will throw exception if invalid
                if('iplowest'===$field){
                    $request[$field]=ip2long($request[$field]);
                    if($request[$field]>$key['iphighest']){
                        throw new BadRequestException('Highest IP cannot be lower than lowest IP');
                    }
                } else if('iphighest'===$field){
                    $request[$field]=ip2long($request[$field]);
                    if($request[$field]<$key['iplowest']){
                        throw new BadRequestException('Lowest IP cannot be higher than highest IP');
                    }
                }
                $params=array(
                    ':id'=>$id,
                    ':newvalue'=>$request[$field]
                );
                database::query('UPDATE `apikey` SET `'.$field.'`=:newvalue WHERE id=:id', $params);
            }
        }
        baseobject::afterUpdateHook($id, 'apikey');
        return array('updated'=>static::getById($id));
    }

    /**
     * @param $id
     * @return array
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws ServerException
     */
    private static function regenerateKey($id){
        $newKey=static::getKeyAndHash();
        database::query(
            'UPDATE apikey SET keyhash=:hash WHERE id=:id',
            array(':hash'=>$newKey['hash'], ':id'=>$id)
        );
        $key=static::getById($id);
        $key['key']=$newKey['key'];
        return $key;
    }

    /**
     * @param array $request
     * @return array
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws ServerException
     */
    public static function create($request = array()){
        session::requireAdmin();
        if(empty($request['iplowest'])){
            $request['iplowest']='0.0.0.0';
        }
        if(empty($request['iphighest'])){
            $request['iphighest']='255.255.255.255';
        }
        $request['iplowest']=ip2long($request['iplowest']);
        if(false===$request['iplowest']){
            throw new BadRequestException('Minimum IP is not valid');
        }
        $request['iphighest']=ip2long($request['iphighest']);
        if(false===$request['iphighest']){
            throw new BadRequestException('Maximum IP is not valid');
        }
        if($request['iphighest'] < $request['iplowest']){
            throw new BadRequestException('Maximum IP is lower than minimum IP');
        }
        baseobject::beforeCreateHook($request,'apikey');
        //generate and hash the key
        $keyAndHash=static::getKeyAndHash();
        $key=$keyAndHash['key'];
        $keyHash=$keyAndHash['hash'];
        database::query(
            'INSERT INTO apikey(keyhash,scope,iplowest,iphighest)
            VALUES(:keyhash,:scope,:iplowest,:iphighest)',
            array(
                ':keyhash'=>$keyHash,
                ':scope'=>$request['scope'],
                ':iplowest'=>$request['iplowest'],
                ':iphighest'=>$request['iphighest']
            )
        );
        $created=apikey::getById(database::getLastInsertId());
        $created['key']=$key;
        baseobject::afterCreateHook($created,'apikey');
        return array(
            'type'=>'apikey',
            'created'=>$created
        );
    }

    /**
     * Deletes the API key.
     * @param $id
     * @return array
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws ServerException
     * @throws NotFoundException
     */
    public static function delete($id){
        session::requireAdmin();
        $apiKey=apikey::getById($id);
        if(!$apiKey){ throw new NotFoundException('No API key with that ID'); }
        baseobject::beforeDeleteHook($apiKey,'apikey');
        database::query(
            'DELETE FROM apikey WHERE id=:id',
            array(':id'=>$id)
        );
        baseobject::afterDeleteHook($apiKey,'apikey');
        return array('deleted'=>$id);
    }

    public static function validate($key, $scope){
        $isValid=false;
        $keys=database::queryGetAll(
            'SELECT * FROM apikey',
            array()
        );
        if(!$keys){
            throw new AuthorizationRequiredException('No API keys defined');
        }
        foreach($keys['rows'] as $row) {
            if(password_verify($key, $row['keyhash'])){
                $ip=ip2long($_SERVER['REMOTE_ADDR']);
                if($ip<(int)$row['iplowest'] || $ip>(int)$row['iphighest']){
                    throw new AuthorizationRequiredException('Originating IP out of range');
                }
                if($scope!=$row['scope']){
                    throw new AuthorizationRequiredException('Scope not authorised for this key');
                }
                $isValid=true;
                break;
            }
        }
        return $isValid;
    }

    private static function getKeyAndHash(){
        $key=bin2hex(random_bytes(30));
        $keyHash=password_hash($key, static::$hashAlgorithm);
        return array('key'=>$key, 'hash'=>$keyHash);
    }

    private static function getScopes(){
        $classNames=[];
        $deviceClasses=rtrim($_SERVER['DOCUMENT_ROOT'], '/').'/classes/device';
        $dir=dir($deviceClasses);
        $filename=$dir->read();
        while(!empty($filename)){
            if(preg_match('/.*\.class\.php$/', $filename)){
                $className=str_replace('.class.php', '', $filename);
                if('Device'!==$className){
                    $classNames[]=$className;
                }

            }
            $filename=$dir->read();
        }
        return $classNames;
    }

}